/**********************************************************************

randclasses.cpp

Class methods for various pseudo random number algorithms.  See
randclasses.h for more info.

**********************************************************************/

#include "randclasses.h"

/**********************************************************************

BlumBlumShub class methods

**********************************************************************/

// public methods

void BlumBlumShub::Init(verylong p, verylong q, verylong key)
{
   /* Initializes the BlumBlumShub random number generator.
      P and Q are both Blum primes greater than 200 digits
      (a Blum prime mod 4 equals 3).
      KEY is a quadratic residue mod (p*q). */

   current = 0;
   modnum = 0;
   zcopy(key,&current);
   zmul(p,q,&modnum);
   NewCurrent();
   NewCurrent();
}

double BlumBlumShub::Random()
{
   double d, *ptr;
   int i;
   unsigned long r[2], u;

   u = 0;
   ptr = (double *)  &r[0];
   for (i = 0; i < ((int) sizeof(unsigned long) * 8); i++)
      u = (u << 1) | (unsigned long) GetBit();
   r[0] = u << 20;
   r[1] = (u >> 12) | 0x3FF00000;
   d = *ptr - 1.0;
   return d;
}

// protected methods

int BlumBlumShub::GetBit()
{
   int the_bit;

   the_bit = (int) zbit(current,pos);
   pos++;
   if (pos = nbits)
      NewCurrent();
   return the_bit;
}

void BlumBlumShub::NewCurrent()
{
   verylong x = 0;

   zsq(current,&x);
   zmod(x,modnum,&current);
   nbits = z2log(current);
   pos = 0;
   zfree(&x);
}

/**********************************************************************

MotherOfAll class methods

**********************************************************************/

// public methods

void MotherOfAll::Init (uint32 seed)
{

   int i;

   // make sure seed != 0

   if (seed==0) seed = -1;

   // make random numbers and put them into the buffer

   for (i=0; i<5; i++)
   {
      seed ^= seed << 13; seed ^= seed >> 17; seed ^= seed << 5;
      x[i] = seed * (1./(65536.*65536.));
   }

   // randomize some more

   for (i=0; i<19; i++) Random();
}

double MotherOfAll::Random()
{

   double c;

   c = 2111111111.0 * x[3] +
       1492.0 * (x[3] = x[2]) +
       1776.0 * (x[2] = x[1]) +
       5115.0 * (x[1] = x[0]) +
       x[4];
   x[4] = floorl(c);
   x[0] = c - x[4];
   x[4] = x[4] * (1./(65536.*65536.));

   return x[0];

}

/**********************************************************************

MTwister class methods

**********************************************************************/

// public methods

void MTwister::Init(uint32 seed)
{
    /* setting initial seeds to mt[N] using
       the generator Line 25 of Table 1 in
       [KNUTH 1981, The Art of Computer Programming
          Vol. 2 (2nd Ed.), pp102]                  */

    mt[0]= seed & 0xffffffff;
    for (mti=1; mti<N; mti++)
        mt[mti] = (69069 * mt[mti-1]) & 0xffffffff;
}

double MTwister::Random()
{
   int kk;
   uint32 y;
   static uint32 mag01[2]={0x0, MATRIX_A};

   // mag01[x] = x * MATRIX_A  for x=0,1

   if (mti >= N) { /* generate N words at one time */

      if (mti == N+1)   /* if sgenrand() has not been called, */
         Init(4357); /* a default initial seed is used   */

      for (kk=0;kk<N-M;kk++)
      {
         y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
         mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1];
      }
      for (;kk<N-1;kk++)
      {
         y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
         mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1];
      }
      y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
      mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1];

      mti = 0;
   }

   y = mt[mti++];
   y ^= TEMPERING_SHIFT_U(y);
   y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
   y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
   y ^= TEMPERING_SHIFT_L(y);

   return ( (double)y / (uint32)0xffffffff );
}


/**********************************************************************

RandC class methods

**********************************************************************/

// public methods

void RandC::Init (uint32 seed)
{
   srand(seed);
}

double RandC::Random()
{
   return ( (double) rand() / (double) (RAND_MAX + 1));
}

/**********************************************************************

Ranshi class methods

**********************************************************************/

// public methods

void Ranshi::Init(uint32 seed)
{
   int i;

   halfBuff = 0;
   numFlats = 0;
   twoToMinus_32 = ldexp (1.0, -32);
   twoToMinus_53 = ldexp (1.0, -53);
   for (i = 0; i < numBuff; ++i)
      buffer[i] = (unsigned int)seed&0xffffffff;
   redSpin = (unsigned int)(seed & 0xffffffff);
   for (i = 0; i < numBuff*20; ++i)        // "warm-up" for engine to hit
      Random();                            //  every ball on average 20X.
}

double Ranshi::Random() {
   unsigned int redAngle = (((numBuff/2) - 1) & redSpin) + halfBuff;
   unsigned int blkSpin     = buffer[redAngle] & 0xffffffff;
   unsigned int boostResult = blkSpin ^ redSpin;

   buffer[redAngle] = ((blkSpin << 17) | (blkSpin >> (32-17))) ^ redSpin;

   redSpin  = (blkSpin + numFlats++) & 0xffffffff;
   halfBuff = numBuff/2 - halfBuff;

   return ( blkSpin * twoToMinus_32 +            // most significant part
            (boostResult>>11) * twoToMinus_53);  // fill in remaining bits
}


/**********************************************************************

RanrotW class methods

**********************************************************************/

// public methods

void RanrotW::Init (uint32 seed)
{

   int i, j;

   // make sure seed != 0

   if (seed==0) seed--;

   // make random numbers and put them into the buffer

   for (i=0; i<KK; i++)
   {
     for (j=0; j<2; j++)
     {
       seed ^= seed << 13; seed ^= seed >> 17; seed ^= seed << 5;
       randbuffer[i][j] = seed;
     }
   }

   // set exponent of randp1

   randp1 = 1.5;

   // initialize pointers to circular buffer

   p1 = 0;  p2 = JJ;

   // randomize some more

   for (i=0; i<97; i++) Random();
}

double RanrotW::Random()
{

  double r;
  uint32 y, z;

  // generate next number

  z = _lrotl(randbuffer[p1][0], R1) + randbuffer[p2][0];
  y = _lrotl(randbuffer[p1][1], R2) + randbuffer[p2][1];
  randbuffer[p1][0] = y;
  randbuffer[p1][1] = z;

  // rotate list pointers

  if (--p1 < 0) p1 = KK - 1;
  if (--p2 < 0) p2 = KK - 1;

  // fast conversion to float:

  randbits[0] = y;
  randbits[1] = (z & 0x000FFFFF) | 0x3FF00000;

  r = randp1 - 1.;
  return r;
}

/**********************************************************************

TripleRand class methods

**********************************************************************/


//********************************************************************
//   TripleRand
//********************************************************************

// public methods

void TripleRand::Init(long seed)
{
   twoToMinus_32 = ldexp (1.0, -32);
   twoToMinus_53 = ldexp (1.0, -53);
   tausworthe.Init((unsigned int)seed + 175321);
   integerCong.Init(69607 * tausworthe.Next() + 54329, 1313);
   hurd.Init(19781127 + integerCong.Next());
}

double TripleRand::Random()
{
   unsigned int ic = integerCong.Next();
   unsigned int t  = tausworthe.Next();
   unsigned int h  = hurd.Next();
   return ( (t ^ ic ^ h) * twoToMinus_32 +          // most significant part
            (h >> 11) * twoToMinus_53
          );
}

// private methods

//********************************************************************
//   Tausworthe
//********************************************************************

void TripleRand::Tausworthe::Init(unsigned int seed)
{
   words[0] = seed;
   for (wordIndex = 1; wordIndex < 4; ++wordIndex)
      words[wordIndex] = 69607 * words[wordIndex-1] + 54329;
}

unsigned int TripleRand::Tausworthe::Next()
{

   if (wordIndex <= 0) {
      for (wordIndex = 0; wordIndex < 4; ++wordIndex)
      {
        words[wordIndex] = ( (words[(wordIndex+1) & 3] << 1 ) |
                                     (words[wordIndex] >> 31)   )
                         ^ ( (words[(wordIndex+1) & 3] << 31) |
                                     (words[wordIndex] >>  1)   );
      }
   }
   return words[--wordIndex] & 0xffffffff;

}

//********************************************************************
//   IntegerCong
//********************************************************************

void TripleRand::IntegerCong::Init(unsigned int seed, int streamNumber)
{
   state = seed;
   multiplier = 65536 + 1024 + 5 + (8 * 1017 * streamNumber);
   addend = 12341;
}

unsigned int TripleRand::IntegerCong::Next()
{
   return state = (state * multiplier + addend) & 0xffffffff;
}

//********************************************************************
//   Hurd
//********************************************************************

void TripleRand::Hurd::Init( long seed )
{
   int i;

   long seedlist[2]={seed,0};
   setSeeds(seedlist, 0);
   words[0] ^= 0xa5482134;        // To make unique vs default two unsigned
   if (words[0]==0) words[0] = 1; // ints in the constructor
   for( i=0; i < 100; ++i ) Next();       // warm up just a bit
}

void TripleRand::Hurd::setSeeds( const long* seeds, int )
{
   setSeed( *seeds ? *seeds : 32767, 0 );
}

void TripleRand::Hurd::setSeed( long seed, int )
{
   words[0] = (unsigned int)seed;
   for (wordIndex = 1; wordIndex < 9; ++wordIndex)
      words[wordIndex] = 69607 * words[wordIndex-1] + 54329;
}

unsigned int TripleRand::Hurd::Next()
{
   register unsigned int W0, W1, W2, W3, W4, W5, W6, W7, W8;

   if( wordIndex <= 1 )
   {
      W0 = words[0];
      W1 = words[1];
      W2 = words[2];
      W3 = words[3];
      W4 = words[4];
      W5 = words[5];
      W6 = words[6];
      W7 = words[7];
      W8 = words[8];
      W1 ^= W0; W0 = f288(W2, W3, W0);
      W2 ^= W1; W1 = f288(W3, W4, W1);
      W3 ^= W2; W2 = f288(W4, W5, W2);
      W4 ^= W3; W3 = f288(W5, W6, W3);
      W5 ^= W4; W4 = f288(W6, W7, W4);
      W6 ^= W5; W5 = f288(W7, W8, W5);
      W7 ^= W6; W6 = f288(W8, W0, W6);
      W8 ^= W7; W7 = f288(W0, W1, W7);
      W0 ^= W8; W8 = f288(W1, W2, W8);
      words[0] = W0 & 0xffffffff;
      words[1] = W1 & 0xffffffff;
      words[2] = W2 & 0xffffffff;
      words[3] = W3 & 0xffffffff;
      words[4] = W4 & 0xffffffff;
      words[5] = W5 & 0xffffffff;
      words[6] = W6 & 0xffffffff;
      words[7] = W7 & 0xffffffff;
      words[8] = W8 & 0xffffffff;
      wordIndex = 9;
   }
   return words[--wordIndex];
}

unsigned int TripleRand::Hurd::f288(unsigned int a, unsigned int b, unsigned int c)
{
  return ( ((b<<2) & 0x7ffc) | ((a<<2) & ~0x7ffc) | (a>>30) ) ^
         ( (c<<1)|(c>>31) );
}

