/**********************************************************************

randclasses.h
Version 1.0 8/12/2000
Mark Cederholm (plp@pierssen.com)

Class declarations for various pseudo random number algorithms.  The
classes are:

BlumBlumShub
MotherOfAll
MTwister
RandC
Ranshi
RanrotW
TripleRand

See individual declarations for more info.

Each class except BlumBlumShub has the following public methods:

Constructor(){};            // does nothing
void Init(uint32 seed);     // initialize the generator
double Random();            // return random n such that 0 <= n < 1

Class BlumBlumShub has the following Init declaration:

void Init(verylong p, verylong q, verylong key);

In all cases, Init must be called before Random.  No safeguards exist to
prevent bad results otherwise.

See individual declarations for more info.

**********************************************************************/

#include <stdlib.h>
#include <math.h>

typedef unsigned long uint32;       // use 32 bit unsigned integers

/**********************************************************************

Class BlumBlumShub is an implementation of the Blum-Blum-Shub algorithm
to generate cryptographically strong random numbers.  The Class is very
slow.

For more info see:

http://www.mit.edu/afs/net.mit.edu/dapps/mit/jis/random/

Init requires verylong P, Q, and KEY.  P and Q are very large, different
Blum prime numbers.  (A Blum prime mod 4 equals 3.)  KEY is a quadratic
residue mod (p*q).

For large integer math, it utilizes the freeLIP package designed by
Arjen K. Lenstra and currently supported by Paul Leyland
[ftp://sable.ox.ac.uk/pub/math/].

**********************************************************************/

#include "lip.h"

class BlumBlumShub
{
   public:

   BlumBlumShub(){};
   void Init(verylong p, verylong q, verylong key);
   double Random();

   protected:

   int GetBit();
   void NewCurrent();
   verylong current;
   verylong modnum;
   long nbits, pos;
};

/**********************************************************************

Class MotherOfAll is taken from the code by Agner Fog [www.agner.org].

**********************************************************************/

class MotherOfAll
{
   public:

   MotherOfAll(){};
   void Init(uint32 seed);
   double Random();

   protected:

   double x[5];
};

/**********************************************************************

Class MTwister is an implementation of the Mersenne Twister algorithm
by M. Matsumoto and T. Nishimura, adapted from the C code at:

http://www.math.keio.ac.jp/~matumoto/emt.html

**********************************************************************/

// Period parameters

#define N 624
#define M 397
#define MATRIX_A 0x9908b0df   // constant vector a
#define UPPER_MASK 0x80000000 // most significant w-r bits
#define LOWER_MASK 0x7fffffff // least significant r bits

// Tempering parameters

#define TEMPERING_MASK_B 0x9d2c5680
#define TEMPERING_MASK_C 0xefc60000
#define TEMPERING_SHIFT_U(y)  (y >> 11)
#define TEMPERING_SHIFT_S(y)  (y << 7)
#define TEMPERING_SHIFT_T(y)  (y << 15)
#define TEMPERING_SHIFT_L(y)  (y >> 18)


class MTwister
{
   public:

   MTwister(){};
   void Init(uint32 seed);
   double Random();

   protected:

   uint32 mt[N];              // the array for the state vector
   int mti;
};

/**********************************************************************

Class RandC uses the standard C++ library rand function

**********************************************************************/

class RandC
{
   public:

   RandC(){};
   void Init(uint32 seed);
   double Random();

};


/**********************************************************************

Class Ranshi is an implementation of the Ranshi algorithm by F. Gutbrod,
adapted from the Fermilab HEP Random library at:

http://www.fnal.gov/docs/working-groups/fpcltf/Pkg/CLHEP/doc/html/CLHEP-random.html

**********************************************************************/

#define numBuff 512

class Ranshi
{
   public:

   Ranshi(){};
   void Init(uint32 seed);
   double Random();

   private:

   double twoToMinus_32;
   double twoToMinus_53;
   unsigned int halfBuff, numFlats;
   unsigned int buffer[numBuff];
   unsigned int redSpin;

};

/**********************************************************************

Class RanrotW is taken from the code by Agner Fog [www.agner.org].

**********************************************************************/

// define parameters

const int KK = 17;
const int JJ = 10;
const int R1 = 19;
const int R2 = 27;

class RanrotW
{
   public:

   RanrotW(){};                    // constructor
   void Init(uint32 seed);         // initialization
   double Random();                // get floating point n between 0 and 1

   protected:

   union                           // used for conversion to float
   {
      double randp1;
      uint32 randbits[2];
   };
   int p1, p2;                     // indexes into buffer
   uint32 randbuffer[KK][2];       // history buffer
};

/**********************************************************************

Class TripleRand is an implementation of the TripleRand algorithm,
adapted from the Fermilab HEP Random library at:

http://www.fnal.gov/docs/working-groups/fpcltf/Pkg/CLHEP/doc/html/CLHEP-random.html

**********************************************************************/

class TripleRand{

   public:

   TripleRand(){};
   void Init( long seed );
   double Random();

   private:

   class Tausworthe
   {
      public:

      Tausworthe(){};
      void Init(unsigned int seed);
      unsigned int Next();

      private:

      int wordIndex;
      unsigned int words[4];
   };

   class IntegerCong
   {
      public:

      IntegerCong(){};
      void Init(unsigned int seed, int streamNumber);
      unsigned int Next();

      private:

      unsigned int state, multiplier, addend;
   };

   class Hurd
   {
      public:

      Hurd(){};
      void Init( long seed );
      unsigned int Next();

      private:

      void setSeeds( const long* seeds, int );
      void setSeed( long seed, int );
      unsigned int f288(unsigned int a, unsigned int b, unsigned int c);

      int wordIndex;
      unsigned int words[9];

   };


  double twoToMinus_32;
  double twoToMinus_53;

  Tausworthe    tausworthe;  // Instances of each of the
  IntegerCong   integerCong; // three engines that combine to make
  Hurd          hurd;        // one TripleRand instance

};

