/**********************************************************************

randlib.h
Version 2.0 8/18/2000
Mark Cederholm (plp@pierssen.com)

Class RandLib encapsulates several pseudo random number algorithms:

BBS BlumBlumShub
MOA MotherOfAll
MTW MTwister
RNC RandC
RSH Ranshi
RRW RanrotW
TPR TripleRand

See randclasses.h for more info on individual algorithms.

Each algorithm has three methods to return numbers:

unsigned char XXX_GetByte()    // returns 0-255
int XXX_GetInteger()           // returns n where min <= n <= max
double XXX_GetDouble()         // returns n where 0.0 <= n < 1.0

Use SetIntegerRange to set min and max for XXX_GetInteger.  Otherwise, 0
will be returned.

For all algorithms except BBS, if the seed has not been explicitly set
by SetSeed, the current time is used.

BBS must be initialized using BBS_Init.  If BBS_Init has not been run,
the numeric methods will return 0. Some additional methods have been
implemented to assist the user in obtaining p, q, and key:

BBS_BlumInit initializes the FreeLIP pseudorandom number generator for
BBS_GenBlumPrime and BBS_GenKey.  SEED is a seed for the pseudorandom
number generator.  If SEED is 0, the current time is used.

BBS_GenBlumPrime generates a pseudorandom, highly probable (50 tests)
Blum prime number for use as P or Q in BBS_Init.  (A Blum prime mod 4
equals 3.)  Returns 0 if failed.  The argument NUMBITS is the required
number of bits for the resulting value.  Run BBS_BlumInit first.

BBS_GenKey generates a quadratic residue mod (P*Q) for use in BBS_Init.
Returns 0 if failed.  Run BBS_BlumInit first.  Adapted from the bc code
by Jeffrey I. Schiller
[http://www.mit.edu/afs/net.mit.edu/dapps/mit/jis/random/].

**********************************************************************/

#include "randclasses.h"
#include <time.h>

class RandLib
{
   public:

   RandLib();

   // algorithm methods

   unsigned char BBS_GetByte();
   int BBS_GetInteger();
   double BBS_GetDouble();
   unsigned char MOA_GetByte();
   int MOA_GetInteger();
   double MOA_GetDouble();
   unsigned char MTW_GetByte();
   int MTW_GetInteger();
   double MTW_GetDouble();
   unsigned char RNC_GetByte();
   int RNC_GetInteger();
   double RNC_GetDouble();
   unsigned char RSH_GetByte();
   int RSH_GetInteger();
   double RSH_GetDouble();
   unsigned char RRW_GetByte();
   int RRW_GetInteger();
   double RRW_GetDouble();
   unsigned char TPR_GetByte();
   int TPR_GetInteger();
   double TPR_GetDouble();

   // initialization

   void SetSeed(uint32 the_seed);
   void SetIntegerRange(int the_min, int the_max);
   void BBS_Init(verylong p, verylong q, verylong key);
   void BBS_BlumInit(verylong seed);
   verylong BBS_GenBlumPrime(int numbits);
   verylong BBS_GenKey(verylong p, verylong q);

   private:

   uint32 GetSeed();
   int AsInteger(double d, int the_min, int the_max);

   BlumBlumShub bbs;
   MotherOfAll moa;
   MTwister mtw;
   RandC rnc;
   Ranshi rsh;
   RanrotW rrw;
   TripleRand tpr;

   int min,max;
   int range_set;
   uint32 seed;

   int bbs_initialized;
   int moa_initialized;
   int mtw_initialized;
   int rnc_initialized;
   int rsh_initialized;
   int rrw_initialized;
   int tpr_initialized;

};


